package http

import (
	"bytes"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"strconv"
	"strings"
	"time"

	"github.com/gogo/protobuf/proto"
	nrpc "github.com/nats-rpc/gogo-nrpc"
	"github.com/rs/zerolog"
	xbus "xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/envelope"
)

var (
	ErrStopIteration = errors.New("stop iteration")
	ErrNoSuchInput   = errors.New("no such input")
)

type WebHookPayloadType string

const (
	// WebHookServiceName is the name of the webhook service in the registry
	WebHookServiceName = "http.webhook"

	WebHookPayloadFullRequest       WebHookPayloadType = "full-request"
	WebHookPayloadEnvelope          WebHookPayloadType = "envelope"
	WebHookPayloadMessageInRequest  WebHookPayloadType = "message-in-request"
	WebHookPayloadMessageInEnvelope WebHookPayloadType = "message-in-envelope"
	WebHookPayloadMessageContent    WebHookPayloadType = "message-content"
)

var ErrInvalidWebHookPayloadType = errors.New("invalid webhook payload type")

func WebHookPayloadTypeFromString(s string) (WebHookPayloadType, error) {
	switch t := WebHookPayloadType(s); t {
	case
		WebHookPayloadFullRequest,
		WebHookPayloadEnvelope,
		WebHookPayloadMessageInRequest,
		WebHookPayloadMessageInEnvelope,
		WebHookPayloadMessageContent:
		return t, nil
	case "":
		return WebHookPayloadFullRequest, nil
	default:
		return WebHookPayloadType(""), fmt.Errorf("%w: %s", ErrInvalidWebHookPayloadType, s)
	}
}

type Header struct {
	Key   string
	Value string
}

// WebHookHandler ...
type WebHookHandler struct {
	url               string
	headers           []Header
	contentType       string
	requestMarshaler  func(r *api.ActorProcessRequest) ([]byte, error)
	envelopeMarshaler func(context.Context, envelope.Envelope) (io.ReadCloser, error)

	payloadType WebHookPayloadType

	successHTTPCodes          []int
	temporaryFailureHTTPCodes []int
}

// NewWebHookHandler ...
func NewWebHookHandler(actor *xbus.Actor) xbus.ActorService {
	url := actor.Options.GetConfigEntry("http.webhook.url")
	if url == "" {
		actor.Log.Errorf("No webhook on this actor")
		return nil
	}
	headers, err := parseHeadersList(
		actor.Options.GetConfigEntry("http.webhook.headers"))
	if err != nil {
		actor.Log.Errorf("Error parsing http.webhook.headers: %s", err)
		return nil
	}
	contentType := actor.Options.GetConfigEntry("http.webhook.content-type")
	if contentType == "" || contentType == "json" {
		contentType = "application/json"
	} else if contentType == "proto" || contentType == "protobuf" {
		contentType = "application/x-protobuf"
	}
	successHTTPCodes, err := parseHTTPCodesList(
		actor.Options.GetConfigEntry("http.webhook.http-codes.success"))
	if err != nil {
		actor.Log.Errorf("Error parsing http.webhook.http-codes.success: %s", err)
		return nil
	}

	temporaryFailureHTTPCodes, err := parseHTTPCodesList(
		actor.Options.GetConfigEntry("http.webhook.http-codes.temporary-failure"))
	if err != nil {
		actor.Log.Errorf("Error parsing http.webhook.http-codes.temporary-failure: %s", err)
		return nil
	}

	payloadType, err := WebHookPayloadTypeFromString(
		actor.Options.GetConfigEntry("http.webhook.payload-type"))
	if err != nil {
		actor.Log.Errorf("Error parsing http.webhook.payload-type: %s", err)
		return nil
	}

	if len(successHTTPCodes) == 0 {
		switch payloadType {
		case WebHookPayloadFullRequest, WebHookPayloadMessageInRequest:
			successHTTPCodes = []int{200}
		case WebHookPayloadEnvelope, WebHookPayloadMessageInEnvelope, WebHookPayloadMessageContent:
			successHTTPCodes = []int{204}
		}
	}

	handler := WebHookHandler{
		url:                       url,
		headers:                   headers,
		contentType:               contentType,
		requestMarshaler:          requestMarshaler(contentType),
		envelopeMarshaler:         envelopeMarshaler(contentType),
		successHTTPCodes:          successHTTPCodes,
		temporaryFailureHTTPCodes: temporaryFailureHTTPCodes,
		payloadType:               payloadType,
	}
	return &handler
}

func requestMarshaler(contentType string) func(r *api.ActorProcessRequest) ([]byte, error) {
	switch contentType {
	case "application/json":
		return jsonMarshalRequest
	case "application/x-protobuf":
		return protobufMarshalRequest
	default:
		panic("no marshaler for contenttype: " + contentType)
	}
}

func jsonMarshalRequest(r *api.ActorProcessRequest) ([]byte, error) {
	bytes, err := json.Marshal(r)
	if err != nil {
		return nil, fmt.Errorf("[http.webhook] Error encoding the payload: %w", err)
	}
	return bytes, nil
}

func protobufMarshalRequest(r *api.ActorProcessRequest) ([]byte, error) {
	bytes, err := proto.Marshal(r)
	if err != nil {
		return nil, fmt.Errorf("[http.webhook] Error encoding the payload: %w", err)
	}
	return bytes, nil
}

func envelopeMarshaler(contentType string) func(context.Context, envelope.Envelope) (io.ReadCloser, error) {
	switch contentType {
	case "application/json":
		return jsonMarshalEnvelope
	case "application/x-protobuf":
		return protobufMarshalEnvelope
	default:
		panic("no marshaler for contenttype: " + contentType)
	}
}

func jsonMarshalEnvelope(ctx context.Context, e envelope.Envelope) (io.ReadCloser, error) {
	complete, err := envelope.CompleteWithContext(ctx, e)
	if err != nil {
		return nil, fmt.Errorf("[http.webhook] failed to read the envelope: %w", err)
	}
	bytes, err := json.Marshal(complete)
	if err != nil {
		return nil, fmt.Errorf("[http.webhook] Error encoding the payload: %w", err)
	}
	r, w := io.Pipe()
	go func() {
		// TODO implement as a stream
		if _, err := w.Write(bytes); err != nil {
			// the error should also be returned to the reader, which should
			// report it. We can ignore it.
			zerolog.Ctx(ctx).Err(err).Msg("[http.webhook] could not write the bytes down the pipe")
		}
		if err := w.Close(); err != nil {
			// the error should also be returned to the reader, which should
			// report it. We can ignore it.
			zerolog.Ctx(ctx).Err(err).Msg("[http.webhook] could not close the pipe")
		}
	}()
	return r, nil
}

func protobufMarshalEnvelope(ctx context.Context, e envelope.Envelope) (io.ReadCloser, error) {
	complete, err := envelope.CompleteWithContext(ctx, e)
	if err != nil {
		return nil, fmt.Errorf("[http.webhook] failed to read the envelope: %w", err)
	}
	bytes, err := proto.Marshal(complete)
	if err != nil {
		return nil, fmt.Errorf("[http.webhook] Error encoding the payload: %w", err)
	}
	r, w := io.Pipe()
	go func() {
		if _, err := w.Write(bytes); err != nil {
			// the error should also be returned to the reader, which should
			// report it. We can ignore it.
			zerolog.Ctx(ctx).Err(err).Msg("[http.webhook] could not write the bytes down the pipe")
		}
		if err := w.Close(); err != nil {
			// the error should also be returned to the reader, which should
			// report it. We can ignore it.
			zerolog.Ctx(ctx).Err(err).Msg("[http.webhook] could not close the pipe")
		}
	}()
	return r, nil
}

func unmarshalActorProcessingState(contentType string, body io.Reader) (api.ActorProcessingState, error) {
	var state api.ActorProcessingState
	switch contentType {
	case "application/json":
		dec := json.NewDecoder(body)
		if err := dec.Decode(&state); err != nil {
			return state, fmt.Errorf("[http.webhook] Error decoding json response: %w", err)
		}
	case "application/x-protobuf":
		b, err := ioutil.ReadAll(body)
		if err != nil {
			return state, fmt.Errorf("[http.webhook] Error reading response: %w", err)
		}
		if err := proto.Unmarshal(b, &state); err != nil {
			return state, fmt.Errorf("[http.webhook] Error decoding protobuf response: %w", err)
		}
	case "":
		return state, fmt.Errorf("[http.webhook] Missing Content-Type header in the response from the webhook")
	default:
		return state, fmt.Errorf("[http.webhook] Invalid Content-Type header in the response from the webhook. Expected 'application/json' or 'application/x-protobuf', got %s", contentType)
	}
	return state, nil
}

func (wh *WebHookHandler) postRequest(
	ctx context.Context, msgtype string, r *api.ActorProcessRequest,
) (api.ActorProcessingState, error) {
	var state api.ActorProcessingState

	url := strings.ReplaceAll(wh.url, "{msgtype}", msgtype)

	payload, err := wh.requestMarshaler(r)
	if err != nil {
		return state, err
	}

	req, err := http.NewRequest("POST", url, bytes.NewBuffer(payload))
	if err != nil {
		return state, fmt.Errorf("failed to forge the request: %w", err)
	}
	req = req.WithContext(ctx)
	req.Header.Set("Content-Type", wh.contentType)
	req.Header.Set("Accept", wh.contentType)
	for _, h := range wh.headers {
		req.Header.Set(h.Key, h.Value)
	}

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return state, &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] %s", err.Error()),
		}
	}
	defer resp.Body.Close()
	if httpCodeMatch(resp.StatusCode, wh.temporaryFailureHTTPCodes) {
		var (
			message = resp.Status
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message += string(buf[:n])
		} else {
			message += ("Could not read body: " + err.Error())
		}
		return state, &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] Temporary failure: %s", message),
		}
	}

	if httpCodeMatch(resp.StatusCode, wh.successHTTPCodes) {
		contentType := resp.Header.Get("Content-Type")
		state, err = unmarshalActorProcessingState(contentType, resp.Body)
		if err != nil {
			state.Status = api.ActorProcessingState_ERROR
			state.Messages = []api.LogMessage{
				{
					Time:  time.Now(),
					Level: api.LogLevel_ERROR,
					Text:  "[http.webhook] " + err.Error(),
				},
			}
			return state, err
		} else if state.Status == api.ActorProcessingState_ERROR && len(state.Messages) == 0 {
			state.Messages = []api.LogMessage{
				{
					Time:  time.Now(),
					Level: api.LogLevel_ERROR,
					Text:  "[http.webhook] The webhook returned a 'ERROR' state without any error message",
				},
			}
		}
	} else {
		state.Status = api.ActorProcessingState_ERROR
		var (
			message string
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message = string(buf[:n])
		} else {
			message = "Could not read body: " + err.Error()
		}
		message = fmt.Sprintf(`[http.webhook] Invalid http status. Expected one of %v, got %s.
The response body is:
%s`,
			wh.successHTTPCodes,
			resp.Status,
			message)
		state.Messages = []api.LogMessage{
			{
				Time:  time.Now(),
				Level: api.LogLevel_ERROR,
				Text:  message,
			},
		}
	}
	return state, nil
}

func (wh *WebHookHandler) postEnvelope(
	ctx context.Context, msgtype string, e envelope.Envelope,
) (api.ActorProcessingState, error) {
	var state api.ActorProcessingState

	url := strings.ReplaceAll(wh.url, "{msgtype}", msgtype)

	payload, err := wh.envelopeMarshaler(ctx, e)
	if err != nil {
		return state, err
	}

	req, err := http.NewRequest("POST", url, payload)
	if err != nil {
		return state, fmt.Errorf("failed to forge the request: %w", err)
	}
	req = req.WithContext(ctx)
	req.Header.Set("Content-Type", wh.contentType)
	req.Header.Set("Accept", wh.contentType)
	for _, h := range wh.headers {
		req.Header.Set(h.Key, h.Value)
	}

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return state, &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] %s", err.Error()),
		}
	}
	defer resp.Body.Close()
	if httpCodeMatch(resp.StatusCode, wh.temporaryFailureHTTPCodes) {
		var (
			message = resp.Status
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message += string(buf[:n])
		} else {
			message += ("Could not read body: " + err.Error())
		}
		return state, &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] Temporary failure: %s", message),
		}
	}

	if httpCodeMatch(resp.StatusCode, wh.successHTTPCodes) {
		state.Status = api.ActorProcessingState_SUCCESS
	} else {
		state.Status = api.ActorProcessingState_ERROR
		var (
			message string
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message = string(buf[:n])
		} else {
			message = "Could not read body: " + err.Error()
		}
		message = fmt.Sprintf(`[http.webhook] Invalid http status. Expected one of %v, got %s.
The response body is:
%s`,
			wh.successHTTPCodes,
			resp.Status,
			message)
		state.Messages = []api.LogMessage{
			{
				Time:  time.Now(),
				Level: api.LogLevel_ERROR,
				Text:  message,
			},
		}
	}
	return state, nil
}

func (wh *WebHookHandler) postContent(
	ctx context.Context, msgtype string, contenttype string, payload io.Reader,
) (api.ActorProcessingState, error) {
	var state api.ActorProcessingState

	url := strings.ReplaceAll(wh.url, "{msgtype}", msgtype)

	req, err := http.NewRequest("POST", url, payload)
	if err != nil {
		return state, fmt.Errorf("failed to forge the request: %w", err)
	}
	req = req.WithContext(ctx)
	req.Header.Set("Content-Type", wh.contentType)
	req.Header.Set("Accept", wh.contentType)
	for _, h := range wh.headers {
		req.Header.Set(h.Key, h.Value)
	}

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return state, &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] %s", err.Error()),
		}
	}
	defer resp.Body.Close()
	if httpCodeMatch(resp.StatusCode, wh.temporaryFailureHTTPCodes) {
		var (
			message = resp.Status
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message += string(buf[:n])
		} else {
			message += ("Could not read body: " + err.Error())
		}
		return state, &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] Temporary failure: %s", message),
		}
	}

	if httpCodeMatch(resp.StatusCode, wh.successHTTPCodes) {
		state.Status = api.ActorProcessingState_SUCCESS
	} else {
		state.Status = api.ActorProcessingState_ERROR
		var (
			message string
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message = string(buf[:n])
		} else {
			message = "Could not read body: " + err.Error()
		}
		message = fmt.Sprintf(`[http.webhook] Invalid http status. Expected one of %v, got %s.
The response body is:
%s`,
			wh.successHTTPCodes,
			resp.Status,
			message)
		state.Messages = []api.LogMessage{
			{
				Time:  time.Now(),
				Level: api.LogLevel_ERROR,
				Text:  message,
			},
		}
	}
	return state, nil
}

// Process a request
func (wh *WebHookHandler) Process(apc *xbus.ActorProcessingContext) error {
	var (
		state    api.ActorProcessingState
		stateErr error
	)
	switch wh.payloadType {
	case WebHookPayloadFullRequest:
		fmt.Println("will post Request")
		for _, input := range apc.Request.Inputs {
			e, err := apc.GetEnvelope(input.Name)
			if err != nil {
				return err
			}
			complete, err := envelope.CompleteWithContext(apc.Context, e)
			if err != nil {
				return err
			}

			input.Envelope = complete
			input.Position = nil
		}
		state, stateErr = wh.postRequest(apc.Context, "", apc.Request)
		fmt.Println("posted Request", state, stateErr)
		if stateErr != nil {
			return stateErr
		}
	case WebHookPayloadEnvelope:
		state.Status = api.ActorProcessingState_SUCCESS
		for _, input := range apc.Request.Inputs {
			e, err := apc.GetEnvelope(input.Name)
			if err != nil {
				return err
			}
			s, err := wh.postEnvelope(apc.Context, "", e)
			if err != nil {
				return err
			}
			if s.Status != api.ActorProcessingState_SUCCESS {
				state = s
				break
			}
			state = s
		}
	case WebHookPayloadMessageInRequest,
		WebHookPayloadMessageInEnvelope,
		WebHookPayloadMessageContent:
		for _, input := range apc.Request.Inputs {
			e, err := apc.GetEnvelope(input.Name)
			if err != nil {
				return err
			}
			if err := envelope.ForEachMessage(e, func(msg envelope.Message) error {
				switch wh.payloadType {
				case WebHookPayloadMessageInRequest:
					request, err := requestWithSingleMessage(
						apc.Context, apc.Request, input.Name, msg)
					if err != nil {
						stateErr = err
					} else {
						state, stateErr = wh.postRequest(apc.Context, msg.MsgType(), request)
					}
				case WebHookPayloadMessageInEnvelope:
					env, err := envelopeWithSingleMessage(apc.Context, e, msg)
					if err != nil {
						stateErr = err
					} else {
						state, stateErr = wh.postEnvelope(apc.Context, msg.MsgType(), env)
					}
				case WebHookPayloadMessageContent:
					state, stateErr = wh.postContent(
						apc.Context, msg.MsgType(), wh.contentType, msg.Reader(),
					)
				}
				// TODO if err is Error_SERVERTOOBUSY on the first message, return it.
				// if Error_SERVERTOOBUSY on a later message, do the wait and retry,
				// after too many tries or on shutdown, return TOOBUSY
				if stateErr != nil {
					state.Status = api.ActorProcessingState_ERROR
					state.Messages = []api.LogMessage{{
						Time:  time.Now(),
						Level: api.LogLevel_ERROR,
						Text:  stateErr.Error(),
					}}
				}
				if state.Status != api.ActorProcessingState_SUCCESS {
					return ErrStopIteration
				}
				return nil
			}); err != nil && !errors.Is(err, ErrStopIteration) {
				return err
			}
		}
	}

	if state.Status == api.ActorProcessingState_PENDING {
		apc.Detach()
	} else {
		apc.End(state.Status, state.Messages)
	}
	return nil
}

func init() {
	xbus.RegisterActorService(WebHookServiceName, NewWebHookHandler)
}

func httpCodeMatch(code int, list []int) bool {
	for _, value := range list {
		if value == code {
			return true
		}
	}
	return false
}

func parseHTTPCodesList(s string) ([]int, error) {
	var (
		splitted = strings.Split(s, ",")
		codes    = make([]int, 0, len(splitted))
	)
	for _, token := range splitted {
		token := strings.TrimSpace(token)
		if token == "" {
			continue
		}
		code, err := strconv.Atoi(token)
		if err != nil {
			return nil, fmt.Errorf("could not parse '%s': %w", token, err)
		}
		codes = append(codes, code)
	}

	return codes, nil
}

func parseHeadersList(s string) ([]Header, error) {
	var (
		splitted = strings.Split(s, "\n")
		res      = make([]Header, 0, len(splitted))
	)
	for _, line := range splitted {
		line = strings.TrimSpace(line)
		if line == "" {
			continue
		}
		sp := strings.SplitN(line, ":", 2)
		if len(sp) < 2 {
			return nil, fmt.Errorf("invalid header line: %s", line)
		}

		res = append(res, Header{
			Key:   strings.TrimSpace(sp[0]),
			Value: strings.TrimSpace(sp[1]),
		})
	}
	return res, nil
}

func requestWithSingleMessage(
	ctx context.Context, r *api.ActorProcessRequest, inputName string, msg envelope.Message,
) (*api.ActorProcessRequest, error) {
	if r == nil {
		panic("nil request")
	}
	var input *api.ActorProcessRequest_Input
	for _, i := range r.Inputs {
		if i.Name == inputName {
			input = i
			break
		}
	}
	if input == nil {
		return nil, fmt.Errorf("%w: input", ErrNoSuchInput)
	}

	chunks, err := envelope.ReadAllChunks(msg.ChunkReader())
	if err != nil {
		return nil, err
	}

	var request = api.ActorProcessRequest{
		Context: r.Context,
		Inputs: []*api.ActorProcessRequest_Input{{
			Name: input.Name,
			Envelope: &api.Envelope{
				ID:       input.Envelope.ID,
				EventIDs: []api.UUID{msg.ID()},
				Events: []api.Event{{
					ID:        msg.ID(),
					Type:      msg.MsgType(),
					Index:     1,
					ItemCount: uint64(len(chunks)),
					Checksum:  msg.Checksum(),
					Items:     api.ChunksAsItems(chunks),
				}},
			},
		}},
	}

	return &request, nil
}

func envelopeWithSingleMessage(
	ctx context.Context, e envelope.Envelope, msg envelope.Message,
) (envelope.Envelope, error) {
	if e == nil {
		panic("nil envelope")
	}
	return envelope.NewEnvelopeWithID(e.ID(), msg), nil
}
