package process

import (
	"bytes"
	"context"
	"fmt"
	"io"
	"time"

	"github.com/rs/zerolog"
	yaml "gopkg.in/yaml.v2"
	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/pipeline"
	"xbus.io/xbus/storage"
)

const envelopeExportChunkSize = 250 * 1024

// EnvelopeActorPair contains a pair (EnvelopeID, ActorID)
type EnvelopeActorPair struct {
	EnvelopeID api.UUID
	ActorID    api.UUID
}

// NewProcess creates a process
func NewProcess(
	actorStorage storage.ActorStorage,
	emitterID,
	envelopeID api.UUID,
	pipeline *pipeline.Pipeline,
	source storage.NodeOutputRef,
	manager *Manager,
) *Process {
	id, err := api.UUIDV4()
	if err != nil {
		panic(fmt.Sprintf("Cannot generate a UUID: %s", err))
	}
	return &Process{
		ID:                id,
		actors:            actorStorage,
		SourceRef:         source,
		TriggerEmitterID:  emitterID,
		Pipeline:          pipeline,
		TriggerEnvelopeID: envelopeID,
		manager:           manager,
	}
}

// Process orchestrate a pipeline execution
type Process struct {
	ID        api.UUID
	CreatedAt time.Time
	GroupID   api.UUID
	ReplayOf  api.UUID

	actors         storage.ActorStorage
	processStorage storage.ProcessStorage
	graphManager   *pipeline.Manager

	manager *Manager

	TriggerEmitterID  api.UUID
	TriggerEnvelopeID api.UUID

	Pipeline  *pipeline.Pipeline
	SourceRef storage.NodeOutputRef
}

// GetStatus returns the current status of the process
func (p Process) GetStatus() api.Process_Status {
	state := p.processStorage.GetState(p.ID)
	return state.Status
}

// Pause switch the process state to ProcessPaused
func (p Process) Pause() error {
	return p.processStorage.SetStatus(p.ID, api.Process_PAUSED, "")
}

// Resume switch the process state to ProcessRunning if previously paused
func (p Process) Resume() error {
	return p.processStorage.SetStatus(p.ID, api.Process_RUNNING, "")
}

// Cancel sets the process state to ERROR with "canceled: <reason>" reason
func (p Process) Cancel(reason string) error {
	return p.processStorage.SetStatus(p.ID, api.Process_ERROR, "Canceled: "+reason)
}

// MatchPipeline attempts to find the pipeline that will drive the process
func (p *Process) MatchPipeline(output string, eventTypes []string) error {
	var success bool
	// TODO UpdateReceiveStatus with ERROR if no match
	defer func() {
		if !success {
			p.manager.Log.Info().Msg("MatchPipeline failed, setting status to error")
			if err := p.processStorage.SetStatus(
				p.ID, api.Process_ERROR, "No matching pipeline",
			); err != nil {
				p.manager.Log.Err(err).Msg("Error setting process status")
			}
		}
	}()
	actor, ok := p.actors.Get(p.TriggerEmitterID)
	if !ok {
		return fmt.Errorf("Invalid actor")
	}
	pipe, outputRef, ok := p.graphManager.LookupPipeline(actor, output, eventTypes)
	if !ok {
		return fmt.Errorf("No matching pipeline")
	}
	p.Pipeline = pipe
	p.SourceRef = outputRef
	p.processStorage.SetPipeline(p.ID, p.Pipeline.ID, p.SourceRef)
	success = true
	return nil
}

// GetTargets returns the targets for an envelope coming from a given emitter
func (p Process) GetTargets(output storage.NodeOutputRef, envelopeID api.UUID, noRouteTableUpdate bool) ([]api.EnvelopeTarget, error) {
	if p.Pipeline == nil {
		return nil, fmt.Errorf("Pipeline-less process")
	}
	// is the envelope already routed ?
	return p.processStorage.GetTargets(p.ID, output, envelopeID, noRouteTableUpdate,
		func() ([]api.EnvelopeTarget, bool, error) {
			dests, err := p.Pipeline.GetDestinations(output)
			if err != nil {
				return nil, false, err
			}
			if len(dests) == 0 {
				panic("found no destination")
			}
			if len(dests) != 1 {
				panic("found more than one destination")
			}
			dest := dests[0]
			destNode, err := p.Pipeline.GetNode(dest.NodeID)
			if err != nil {
				return nil, false, fmt.Errorf("cannot find the target node: %w", err)
			}

			var actors []api.Actor

			switch destNode.Type {
			case storage.ConsumerNode:
				if len(destNode.Roles) != 0 && len(destNode.Actors) != 0 {
					return nil, false, fmt.Errorf(
						"consumer %s node has both roles and actors defined",
						destNode.ID,
					)
				}

				switch {
				case len(destNode.Roles) >= 1:
					if !destNode.RoleBroadcast || len(destNode.Roles) != 1 {
						panic("not implemented")
					}
					actors = p.actors.ListRole(api.Actor_CONSUMER, destNode.Roles[0])
				case len(destNode.Actors) >= 1:
					for _, actorRef := range destNode.Actors {
						actorID := api.MaybeUUIDFromString(actorRef)
						if !actorID.IsZero() {
							actor, exists := p.actors.Get(actorID)
							if !exists {
								return nil, false, fmt.Errorf("No such actor: %s", actorID)
							}
							actors = append(actors, actor)
						} else {
							actor, exists := p.actors.GetByName(actorRef)
							if !exists {
								return nil, false, fmt.Errorf("No such actor: %s", actorRef)
							}
							actors = append(actors, actor)
						}
					}
				default:
					panic("not implemented")
				}
			case storage.WorkerNode:
				if destNode.RoleBroadcast || len(destNode.Roles) != 0 {
					panic("Worker graph nodes supports only Actors, not roles")
				}

				for _, actorID := range destNode.ActorIDs {
					actor, exists := p.actors.Get(actorID)
					if !exists {
						return nil, false, fmt.Errorf("No such actor: %s", actorID)
					}
					actors = append(actors, actor)
				}

				for _, actorRef := range destNode.Actors {
					actorID := api.MaybeUUIDFromString(actorRef)
					if !actorID.IsZero() {
						actor, exists := p.actors.Get(actorID)
						if !exists {
							return nil, false, fmt.Errorf("No such actor: %s", actorID)
						}
						actors = append(actors, actor)
					} else {
						actor, exists := p.actors.GetByName(actorRef)
						if !exists {
							return nil, false, fmt.Errorf("No such actor: %s", actorRef)
						}
						actors = append(actors, actor)
					}
				}
			case storage.RequestorNode:
				actor, ok := p.actors.Get(p.TriggerEmitterID)
				if !ok {
					panic("The requestor disappeared during process. Please report")
				}
				actors = append(actors, actor)
			}

			if len(actors) == 0 {
				return nil, false, fmt.Errorf("No target actor found")
			}
			var targets []api.EnvelopeTarget
			for _, actor := range actors {
				targets = append(targets, api.EnvelopeTarget{
					ActorID: actor.ID,
					NodeID:  dest.NodeID,
					Input:   dest.Input,
				})
			}

			isResponse := destNode.Type == storage.RequestorNode && dest.Input == "response"
			return targets, isResponse, nil
		},
	)
}

// UpdateReceiveStatus is called whenever an envelope fragment is received from a node
func (p *Process) UpdateReceiveStatus(envelopeID, actorID api.UUID, outputref storage.NodeOutputRef, status storage.ReceptionStatus) error {
	return p.processStorage.UpdateReceiveStatus(
		storage.Reception{
			ProcessID:       p.ID,
			EnvelopeID:      envelopeID,
			SourceActor:     actorID,
			SourceOutputRef: outputref,
		},
		status)
}

// ActorProcessingEnd informs the process
func (p *Process) ActorProcessingEnd(actorID api.UUID, nodeID string, status api.ActorProcessingState_Status, log zerolog.Logger) {
	if err := p.processStorage.ProcessNodeEnd(
		p.ID, storage.ProcessNode{ActorID: actorID, NodeID: nodeID}, status); err != nil {
		log.Err(err).Msg("Error calling ProcessNodeEnd")
	}
}

// GetResponse returns the process response envelope id if any
func (p *Process) GetResponse() api.UUID {
	state := p.processStorage.GetState(p.ID)
	return state.ResponseEnvelopeID
}

// GetPostMortemState set the postmortem state of the process
func (p *Process) GetPostMortemState() (storage.PostMortemState, error) {
	if status := p.GetStatus(); status != api.Process_DONE && status != api.Process_ERROR {
		return storage.PostMortemState{}, fmt.Errorf("Cannot get postmortem state on a non-ended process. Current status is '%s'", status)
	}
	return p.processStorage.GetPostMortemState(p.ID), nil
}

// SetPostMortemState set the postmortem state of the process
func (p *Process) SetPostMortemState(state storage.PostMortemState) error {
	if status := p.GetStatus(); status != api.Process_DONE && status != api.Process_ERROR {
		return fmt.Errorf("Cannot set postmortem state on a non-ended process. Current status is '%s'", status)
	}
	p.processStorage.SetPostMortemState(p.ID, state)
	return nil
}

func indentAndSend(buf, pad []byte, out io.Writer) error {
	lines := bytes.Split(buf, []byte("\n"))
	for _, line := range lines {
		if _, err := out.Write(pad); err != nil {
			return err
		}
		if _, err := out.Write(line); err != nil {
			return err
		}
		if _, err := out.Write([]byte("\n")); err != nil {
			return err
		}
	}
	return nil
}

// ExportStream export the process as a stream, which supports big envelopes
// and logs
func (p *Process) ExportStream(
	ctx context.Context,
	envelopeStorage storage.EnvelopeStorage,
	jobStorage storage.JobStorage,
	logStorage storage.LogStorage,
	log zerolog.Logger,
	out io.Writer,
	exportEnvelope bool,
	exportJobs bool,
	exportLogs bool,
	exportProcessLogs bool,
	envelopeSizeLimit int64,
	logLevel api.LogLevel,
) error {
	var export = make(map[string]interface{})
	export["id"] = p.ID
	if !p.GroupID.IsZero() {
		export["group_id"] = p.GroupID
	}
	if !p.ReplayOf.IsZero() {
		export["replay_of"] = p.ReplayOf
	}
	export["created_at"] = p.CreatedAt.String()

	emitterEx := map[string]interface{}{
		"id": p.TriggerEmitterID,
	}

	if emitter, exists := p.actors.Get(p.TriggerEmitterID); exists {
		emitterEx["name"] = emitter.Name
	}

	export["trigger"] = map[string]interface{}{
		"emitter":  emitterEx,
		"envelope": p.TriggerEnvelopeID,
	}

	export["pipeline"] = p.Pipeline

	export["status"] = p.GetStatus()
	if p.GetStatus() == api.Process_DONE || p.GetStatus() == api.Process_ERROR {
		pmState, err := p.GetPostMortemState()
		if err != nil {
			panic(err)
		}
		export["postmortem_state"] = pmState
	}

	export["state"] = p.processStorage.GetState(p.ID)

	b, err := yaml.Marshal([]interface{}{export})
	if err != nil {
		return err
	}

	if _, err := out.Write(b); err != nil {
		return err
	}

	if exportJobs {
		jobs := jobStorage.ListJobByProcess(p.ID)
		b, err := yaml.Marshal(map[string][]storage.Job{
			"jobs": jobs,
		})

		if err != nil {
			panic(err)
		}

		if err := indentAndSend(b, []byte("  "), out); err != nil {
			return err
		}
	}

	if exportLogs {
		logs := logStorage.GetProcessMessages(p.ID, logLevel)

		b, err = yaml.Marshal(map[string]interface{}{
			"logs": logs,
		})

		if err != nil {
			panic(err)
		}

		if err := indentAndSend(b, []byte("  "), out); err != nil {
			return err
		}
	}

	var logs []storage.ProcessLogEntry
	if exportProcessLogs || exportEnvelope {
		logs = p.manager.ProcessStorage.GetProcessLogEntries(p.ID)
	}

	if exportProcessLogs {
		b, err = yaml.Marshal(map[string]interface{}{
			"processlogs": logs,
		})

		if err != nil {
			panic(err)
		}

		if err := indentAndSend(b, []byte("  "), out); err != nil {
			return err
		}
	}

	if exportEnvelope {
		if _, err := out.Write([]byte("  envelopes:\n")); err != nil {
			return err
		}
		envelopeIDs := []api.UUID{p.TriggerEnvelopeID}
		envelopeIDset := map[api.UUID]bool{p.TriggerEnvelopeID: true}

		for _, log := range logs {
			if log.Emission != nil {
				if _, ok := envelopeIDset[log.Emission.EnvelopeID]; !ok {
					envelopeIDs = append(envelopeIDs, log.Emission.EnvelopeID)
					envelopeIDset[log.Emission.EnvelopeID] = true
				}
			}
		}

		var size int64

		for _, id := range envelopeIDs {
			if ctx.Err() != nil {
				return ctx.Err()
			}
			envelope, pos, err := envelopeStorage.ReadEnvelope(
				id, storage.EnvelopeStartPos(), envelopeExportChunkSize)
			if err != nil {
				log.Err(err).
					Str("envelope-id", id.String()).
					Msg("cannot read envelope")
				continue
			}
			size += envelopeExportChunkSize
			var b []byte
			if pos.Complete {
				b, err = yaml.Marshal(map[string]interface{}{
					id.String(): envelope,
				})
			} else {
				b, err = yaml.Marshal(map[string][]interface{}{
					id.String(): {envelope},
				})
			}
			if err != nil {
				panic(err)
			}
			if ctx.Err() != nil {
				return ctx.Err()
			}
			if err := indentAndSend(b, []byte("    "), out); err != nil {
				return err
			}

			lastPos := pos
			for !pos.Complete && (envelopeSizeLimit == 0 || size < envelopeSizeLimit) {
				envelope, pos, err = envelopeStorage.ReadEnvelope(
					id, lastPos, envelopeExportChunkSize,
				)
				if err != nil {
					log.Err(err).Str("envelope_id", id.String()).
						Msg("cannot read envelope")
					break
				}
				size += envelopeExportChunkSize
				if pos.Equals(lastPos) {
					log.Warn().Str("envelope-id", id.String()).
						Msg("envelope is not complete")
					break
				}
				b, err = yaml.Marshal([]interface{}{envelope})
				if err != nil {
					panic(err)
				}
				if ctx.Err() != nil {
					return ctx.Err()
				}
				if err := indentAndSend(b, []byte("    "), out); err != nil {
					return err
				}

				lastPos = pos
			}
		}
	}

	return nil
}

// Export exports all the data of the process, including envelopes and logs,
// as a map suitable for json or yaml export
// This implementation is not big-envelope friendly and is now deprecated.
// See ExportStream
func (p *Process) Export(envelopeStorage storage.EnvelopeStorage, log zerolog.Logger) (map[string]interface{}, error) {
	var export = make(map[string]interface{})
	export["id"] = p.ID

	emitterEx := map[string]interface{}{
		"id": p.TriggerEmitterID,
	}

	if emitter, exists := p.actors.Get(p.TriggerEmitterID); exists {
		emitterEx["name"] = emitter.Name
	}

	export["trigger"] = map[string]interface{}{
		"emitter":  emitterEx,
		"envelope": p.TriggerEnvelopeID,
	}

	export["pipeline"] = p.Pipeline

	export["status"] = p.GetStatus()
	if p.GetStatus() == api.Process_DONE || p.GetStatus() == api.Process_ERROR {
		pmState, err := p.GetPostMortemState()
		if err != nil {
			panic(err)
		}
		export["postmortem_state"] = pmState
	}

	export["state"] = p.processStorage.GetState(p.ID)

	envelopeIDs := []api.UUID{p.TriggerEnvelopeID}

	var envelopes = make([]api.Envelope, 0, len(envelopeIDs))

	for _, id := range envelopeIDs {
		envelope, pos, err := envelopeStorage.ReadEnvelope(
			id, storage.EnvelopeStartPos(), 100*1024*1024)
		if err != nil {
			log.Err(err).
				Str("envelope-id", id.String()).
				Msg("cannot read envelope")
			continue
		}
		if !pos.Complete {
			log.Warn().
				Str("envelope-id", id.String()).
				Msg("envelope is too big (>100Ko)")
			continue
		}
		envelopes = append(envelopes, envelope)
	}

	export["envelopes"] = envelopes

	return export, nil
}
