// Copyright 2022 Orus.io. All rights reserved.

package control

import (
	"fmt"
	"os"
	"runtime"
	"runtime/pprof"
	"runtime/trace"
	"strconv"
	"strings"
	"time"
)

var pprofDebugInfoEntry = DebugInfoEntry{
	Name:        "pprof",
	Description: "Return a pprof profile",
	Help: `usage: pprof -- <profile> [--debug] [--gc]

available profiles:

goroutine    - stack traces of all current goroutines
heap         - a sampling of memory allocations of live objects
allocs       - a sampling of all past memory allocations
threadcreate - stack traces that led to the creation of new OS threads
block        - stack traces that led to blocking on synchronization primitives
mutex        - stack traces of holders of contended mutexes
cmdline      - the running program's command line, with arguments separated by NUL bytes.
profile      - the pprof-formatted cpu profile (default --seconds=30s)
trace        - the execution trace in binary form (default --seconds=1s).



options:

--debug    Writes the legacy text format with comments translating addresses
           to function names and line numbers, so that a programmer can read the
           profile without tools.
--gc       Run a Garbage Collection prior to writing the report. Makes sense
           with the 'heap' profile only.
--seconds  Set how many seconds a profiling will be done.

See also: https://pkg.go.dev/runtime/pprof#Profile
`,
	Run: func(args []string, formatter OutputFormatter) error {
		var (
			debug   = 0
			gc      = false
			seconds float64
			profile = ""
		)
		for _, arg := range args {
			switch {
			case arg == "--debug":
				debug = 1
			case arg == "--gc":
				gc = true
			case strings.HasPrefix(arg, "--seconds="):
				i, err := strconv.ParseFloat(arg[10:], 64)
				if err != nil {
					return fmt.Errorf("--seconds expected an integer: %w", err)
				}
				seconds = i
			default:
				if profile != "" {
					return fmt.Errorf("cannot get multiple profiles")
				}
				profile = arg
			}
		}
		switch profile {
		case "":
			return fmt.Errorf("no profile selected")
		case "cmdline":
			sep := "\x00"
			if debug == 1 {
				sep = " "
			}
			fmt.Fprint(formatter.Writer(), strings.Join(os.Args, sep))
			if debug == 1 {
				fmt.Fprint(formatter.Writer(), "\n")
			}
		case "profile":
			if seconds == 0 {
				seconds = 30
			}
			if err := pprof.StartCPUProfile(formatter.Writer()); err != nil {
				// StartCPUProfile failed, so no writes yet.
				return fmt.Errorf("Could not enable CPU profiling: %w", err)
			}
			defer pprof.StopCPUProfile()
			time.Sleep(time.Duration(seconds) * time.Second)
		case "trace":
			if seconds == 0 {
				seconds = 1
			}

			if err := trace.Start(formatter.Writer()); err != nil {
				// trace.Start failed, so no writes yet.
				return fmt.Errorf("Could not enable tracing: %w", err)
			}
			time.Sleep(time.Duration(seconds * float64(time.Second)))
			trace.Stop()
		default:
			p := pprof.Lookup(profile)
			if p == nil {
				return fmt.Errorf("no such profile: %s", profile)
			}
			if gc {
				runtime.GC()
			}
			if err := p.WriteTo(formatter.Writer(), debug); err != nil {
				return err
			}
		}
		return nil
	},
}

func init() {
	RegisterDebugInfoEntry(&pprofDebugInfoEntry)
}
