package director

import (
	"context"
	"fmt"

	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/storage"
)

type directorServer struct {
	*Service
}

func (s directorServer) Log(
	ctx context.Context, actorSID string, req api.ActorLogRequest,
) (err error) {
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return
	}

	if req.EnvelopeID != nil && !req.EnvelopeID.IsZero() {
		_, err = s.Opts.ProcessManager.Get(req.Context.ProcessID)
		if err != nil {
			return
		}
	}

	var (
		envID, processID api.UUID
		nodeID           string
	)
	if req.EnvelopeID != nil {
		envID = *req.EnvelopeID
	}
	if ctx := req.Context; ctx != nil {
		nodeID = ctx.NodeID
		processID = ctx.ProcessID
	}
	for _, msg := range req.Messages {
		s.Opts.Logger.Info().Interface("msg", msg).Msg("logging")
		s.Opts.LogStorage.Log(api.LogEntry{
			EnvelopeID: envID,
			ActorID:    actorID,
			NodeID:     nodeID,
			ProcessID:  processID,
			Message:    msg,
		})
	}
	return
}

func (s directorServer) ProcessingEnd(
	ctx context.Context, actorSID string, req api.ActorProcessingState,
) (err error) {
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return
	}

	p, err := s.Opts.ProcessManager.Get(req.Context.ProcessID)
	if err != nil {
		return
	}

	if req.Status != api.ActorProcessingState_SUCCESS &&
		req.Status != api.ActorProcessingState_ERROR {
		return fmt.Errorf("Invalid status: %s", req.Status)
	}

	for _, msg := range req.Messages {
		s.Opts.LogStorage.Log(api.LogEntry{
			ActorID:   actorID,
			NodeID:    req.Context.NodeID,
			ProcessID: req.Context.ProcessID,
			Message:   msg,
		})
	}

	s.Opts.JobStorage.SetState(storage.JobID{
		ProcessID: req.Context.ProcessID,
		Node: storage.ProcessNode{
			ActorID: actorID,
			NodeID:  req.Context.NodeID,
		},
	}, storage.JobState{
		Detached: true,
		Status:   req.Status,
	})
	p.ActorProcessingEnd(actorID, req.Context.NodeID, req.Status, s.Opts.Logger)
	return
}

func (s directorServer) GetRunningActorRequestList(
	ctx context.Context, actorSID string,
	send func(api.ActorProcessRequest),
) error {
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return err
	}
	jobList, err := s.Opts.JobStorage.ListDetachedJobByActor(actorID)
	if err != nil {
		return err
	}
	defer jobList.Close()
	for jobList.Next() && ctx.Err() == nil {
		job, err := jobList.Get()
		if err != nil {
			return err
		}
		req := api.ActorProcessRequest{
			Context: &api.ProcessingContext{
				ProcessID: job.ProcessID,
				NodeID:    job.Node.NodeID,
			},
		}

		for _, input := range s.processManager.ProcessStorage.GetInputs(
			job.ProcessID, job.Node,
		) {
			envelope, pos, err := s.Opts.EnvelopeStorage.ReadEnvelope(
				input.EnvelopeID, storage.EnvelopeStartPos(), 100*1024)
			if err != nil {
				s.Service.Log.Error().
					Str("envelope-id", input.EnvelopeID.String()).
					Msg("could not read envelope")
				return err
			}
			req.Inputs = append(req.Inputs, &api.ActorProcessRequest_Input{
				Name:     input.Name,
				Envelope: &envelope,
				Position: storage.EncodePosition(envelope.ID, pos),
			},
			)
		}
		if ctx.Err() == nil {
			send(req)
		}
	}
	return ctx.Err()
}
