// Copyright 2016 Orus.io. All rights reserved.

package director

import (
	"context"
	"time"

	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/process"
	"xbus.io/xbus/services"
	svc_tools "xbus.io/xbus/services/tools"
	"xbus.io/xbus/storage"
)

const (
	// DirectorQueue is the name we use to queueSub to a queue with the director
	DirectorQueue = "director_queue"
)

// Options is used to create a Service
type Options struct {
	services.Options

	ActorStorage    storage.ActorStorage
	EnvelopeStorage storage.EnvelopeStorage
	LogStorage      storage.LogStorage
	ProcessManager  *process.Manager
	JobStorage      storage.JobStorage
}

// NewService returns a new Service
func (o Options) NewService() *Service {
	return &Service{
		BaseService:    services.NewBaseService("director", o.Options),
		Opts:           o,
		processManager: o.ProcessManager,
		envWatcher:     NewEnvelopeWatcher(),
	}
}

// New returns a new Service
func New(options Options) *Service {
	return options.NewService()
}

// Job ...
type Job struct {
	storage.Job
	storage storage.JobStorage

	timeoutCount uint

	// If currently running, cancel will stop the job
	cancel func()

	// If currently waiting for a new attempt, unschedule the job
	softCancel func()
}

func (job *Job) SaveState() {
	job.storage.SetState(job.JobID, job.JobState)
}

type CancelProcessJobsArgs struct {
	ProcessID api.UUID
	Soft      bool
}

// Service is the director service implementation
// The director service is responsible for calling the 'Process' function
// on the actors.
// It monitors the ProcessNodeReadyEvent events and schedule jobs
// When a actor is back online or a when a process is resumed, the corresponding
// jobs are resheduled
// The director service acts alone a is not (yet) able to cooperate with other
// director instances.
type Service struct {
	services.BaseService
	Opts Options

	processManager *process.Manager

	api               *api.Client
	directorHandler   *api.DirectorHandler
	actorMap          *svc_tools.ActorMap
	jobs              chan Job
	cancelProcessJobs chan CancelProcessJobsArgs

	envWatcher *EnvelopeWatcher

	ctx    context.Context
	cancel func()

	done chan struct{}
}

func (s *Service) mainloop() {
	defer s.SafeTaskDone()

	actorSet := newActorSet(s.ctx, ActorSetConfig{
		Log:             s.Log,
		Conn:            s.Conn,
		ActorStorage:    s.Opts.ActorStorage,
		JobStorage:      s.Opts.JobStorage,
		ProcessManager:  s.Opts.ProcessManager,
		EnvelopeStorage: s.Opts.EnvelopeStorage,
		ActorMap:        s.actorMap,
		Director:        directorServer{s},
		EnvelopeWatcher: s.envWatcher,
	})

	defer actorSet.Close()

	for _, id := range s.actorMap.ListOnline() {
		actorSet.SetOnline(id)
	}

	for {
		select {
		case job := <-s.jobs:
			actorSet.Submit(&job)

		case event := <-s.actorMap.Events:
			if event.Online {
				actorSet.SetOnline(event.ID)
			} else {
				actorSet.SetOffline(event.ID)
			}

		case args := <-s.cancelProcessJobs:
			actorSet.CancelProcessJobs(args.ProcessID, args.Soft)

		case <-s.done:
			return
		}
	}
}

// Startup initialize all the subscriptions
func (s *Service) Startup() error {
	s.ctx, s.cancel = context.WithCancel(context.Background())
	s.directorHandler = api.NewDirectorHandler(
		s.ctx, s.Conn, directorServer{s},
	)
	if err := s.SubscribeAll(
		s.HandlerQueueSubscribe(s.directorHandler, DirectorQueue),
	); err != nil {
		return err
	}

	s.done = make(chan struct{})
	s.jobs = make(chan Job)
	s.cancelProcessJobs = make(chan CancelProcessJobsArgs)

	s.api = api.NewClient(s.Conn)

	sub, err := s.api.CoreEvents.ProcessNodeReadySubscribe(
		s.handleProcessNodeReady)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)

	sub, err = s.api.CoreEvents.ProcessSubscribe(s.handleProcess)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)

	if m, err := svc_tools.NewActorMap(s.ctx, s.api, s.Log); err == nil {
		s.actorMap = m
	} else {
		return err
	}

	sub, err = s.api.CoreEvents.EnvelopeSubscribe(s.envWatcher.PushEvent)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)

	s.TaskAdd()
	go s.mainloop()

	return nil
}

// Shutdown the service
func (s *Service) Shutdown() error {
	if err := s.UnsubscribeAll(); err != nil {
		s.Log.Err(err).Msg("UnsubscribeAll failed")
	}
	s.actorMap.Close()
	close(s.done)
	t := time.AfterFunc(5*time.Second, func() {
		s.Log.Info().Msg("Forcing shutdown of current calls")
		s.cancel()
	})
	err := s.BaseService.Shutdown()
	t.Stop()
	close(s.jobs)
	s.jobs = nil
	return err
}

func (s *Service) handleProcessNodeReady(event api.ProcessNodeReadyEvent) {
	sJob := storage.Job{
		JobID: storage.JobID{
			ProcessID: event.ProcessID,
			Node: storage.ProcessNode{
				ActorID: event.ActorID,
				NodeID:  event.NodeID,
			},
		},
		JobState: storage.JobState{
			CreatedAt: time.Now(),
		},
	}
	s.Opts.JobStorage.NewJob(sJob.JobID, sJob.JobState)

	s.jobs <- Job{Job: sJob, storage: s.Opts.JobStorage}
}

func (s *Service) handleProcess(event api.CoreEvent) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	if event.Name == "status" {
		switch {
		case event.ProcessOldStatus == api.Process_PAUSED &&
			event.Process.Status == api.Process_RUNNING:
			for _, j := range s.Opts.JobStorage.ListJobByProcess(event.Process.ID) {
				s.Log.Info().Str("node", j.Node.String()).
					Msg("rescheduling processings")
				s.jobs <- Job{Job: j, storage: s.Opts.JobStorage}
			}
		case event.Process.Status == api.Process_PAUSED:
			// we need to cancel ongoing processing if any
			s.cancelProcessJobs <- CancelProcessJobsArgs{
				ProcessID: event.Process.ID,
				Soft:      true,
			}
		case event.Process.Status == api.Process_ERROR:
			// we need to cancel ongoing processing if any
			s.cancelProcessJobs <- CancelProcessJobsArgs{
				ProcessID: event.Process.ID,
				Soft:      false,
			}
		}
	}
}
