package cache

import (
	"sort"
	"sync"
	"sync/atomic"
	"time"

	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/api/control"
	"xbus.io/go-xbus/v3/tools"

	"xbus.io/xbus/storage"
)

var (
	DefaultEnvelopeCacheMaxSize = int64(1024 * 1024 * 1024)
)

type EnvelopeCacheEntry struct {
	aTime time.Time

	eventTypes []string

	validator tools.EnvelopeValidator

	fragmentList []EnvelopeCacheEntryFragment
	size         int64
}

type EnvelopeCacheEntryFragment struct {
	pos      storage.EnvelopeReadPos
	nextPos  storage.EnvelopeReadPos
	size     int64
	fragment *api.Envelope
}

// NewEnvelopeStorage creates a cached envelope storage.
func NewEnvelopeStorage(backend storage.EnvelopeStorage, maxsize int64, log zerolog.Logger) *EnvelopeStorage {
	if maxsize == 0 {
		maxsize = DefaultEnvelopeCacheMaxSize
	}
	sizeInBytesMaxGauge.WithLabelValues("envelope", "default").Set(float64(maxsize))

	return &EnvelopeStorage{
		backend:    backend,
		maxsize:    maxsize,
		envmaxsize: maxsize / 3,
		log:        log,

		entriesMap: make(map[api.UUID]*EnvelopeCacheEntry),
	}
}

// EnvelopeStorage implements a cache layer on top of a storage.EnvelopeStorage
// The cache is populated during store, NOT during read.
// As such it is optimized for a write followed by consecutive reads on the
// same instance, which should cover most the needs until we do actual load-balancing
type EnvelopeStorage struct {
	backend    storage.EnvelopeStorage
	maxsize    int64
	envmaxsize int64
	log        zerolog.Logger
	// TODO replace with a double-link list to get constant time put-in-front on access
	entries    []*EnvelopeCacheEntry
	entriesMap map[api.UUID]*EnvelopeCacheEntry
	entriesM   sync.Mutex
	size       int64
}

func (s *EnvelopeStorage) SetEventHandler(handler storage.EnvelopeStorageEventHandler) {
	s.backend.SetEventHandler(handler)
}

func (s *EnvelopeStorage) CheckStalled(id api.UUID, duration time.Duration) bool {
	return s.backend.CheckStalled(id, duration)
}

func (s *EnvelopeStorage) ReceptionStatus(id api.UUID) (storage.EnvelopeStorageReceptionStatus, error) {
	return s.backend.ReceptionStatus(id)
}

func (s *EnvelopeStorage) Purge(filter storage.EnvelopeFilter) {
	s.backend.Purge(filter)
}

func (s *EnvelopeStorage) Stat() ([]control.StorageStat_Entry, error) {
	stats, err := s.backend.Stat()
	if err != nil {
		return nil, err
	}
	var name string
	for i, s := range stats {
		if i != 0 {
			name += ", "
		}
		name += s.Name
	}
	s.entriesM.Lock()
	stats = append(stats, control.StorageStat_Entry{
		Name:           "cache(" + name + ")",
		Implementation: "cache",
		Count:          uint64(len(s.entries)),
		EstimatedSize:  uint64(atomic.LoadInt64(&s.size)),
	})
	s.entriesM.Unlock()
	return stats, nil
}

func (s *EnvelopeStorage) getCacheEntry(id api.UUID) *EnvelopeCacheEntry {
	s.entriesM.Lock()
	entry := s.entriesMap[id]
	if entry != nil {
		entry.aTime = time.Now()
	}
	s.entriesM.Unlock()
	return entry
}

func (s *EnvelopeStorage) mustCacheEntry(id api.UUID) *EnvelopeCacheEntry {
	s.entriesM.Lock()
	if entry, ok := s.entriesMap[id]; ok {
		s.entriesM.Unlock()
		return entry
	}
	entry := &EnvelopeCacheEntry{
		aTime:     time.Now(),
		validator: tools.NewEnvelopeValidator(),
		size:      48, // seems to be the size of an empty validator
	}
	s.entries = append(s.entries, entry)
	s.entriesMap[id] = entry
	newsize := atomic.AddInt64(&s.size, entry.size)
	sizeInBytesGauge.WithLabelValues("envelope", "default").Set(float64(newsize))
	sizeGauge.WithLabelValues("envelope", "default").Set(float64(len(s.entries)))
	s.entriesM.Unlock()
	return entry
}

func (s *EnvelopeStorage) GetEventTypes(id api.UUID) (types []string, exists bool) {
	if entry := s.getCacheEntry(id); entry != nil && len(entry.eventTypes) != 0 {
		return entry.eventTypes, true
	}
	return s.backend.GetEventTypes(id)
}

func (s *EnvelopeStorage) StoreEnvelope(envelope api.Envelope) (storage.EnvelopeState, error) {
	state, err := s.backend.StoreEnvelope(envelope)
	if err != nil {
		return state, err
	}
	entry := s.mustCacheEntry(envelope.ID)
	if entry.validator.IsNext(envelope) && entry.size+int64(envelope.Size()) < s.envmaxsize {
		pos := storage.EnvelopeStartPos()
		if len(entry.fragmentList) != 0 {
			pos = entry.fragmentList[len(entry.fragmentList)-1].nextPos
		}
		if err := entry.validator.Add(envelope); err != nil {
			s.log.Err(err).Msg("cannot add fragment to the validator")
			return state, nil
		}
		nextPos := estateToPos(state)
		if !nextPos.Equals(pos) {
			fEntry := EnvelopeCacheEntryFragment{
				pos:      pos,
				nextPos:  nextPos,
				size:     int64(envelope.Size()),
				fragment: &envelope,
			}
			entry.fragmentList = append(entry.fragmentList, fEntry)
			atomic.AddInt64(&entry.size, fEntry.size)
			newsize := atomic.AddInt64(&s.size, fEntry.size)
			sizeInBytesGauge.WithLabelValues("envelope", "default").Set(float64(newsize))
			if newsize >= s.maxsize {
				s.needCleanup()
			}
		}
	}
	return state, nil
}

func (s *EnvelopeStorage) ReadEnvelope(id api.UUID, position storage.EnvelopeReadPos, maxsize int) (api.Envelope, storage.EnvelopeReadPos, error) {
	entry := s.mustCacheEntry(id)
	for _, fentry := range entry.fragmentList {
		if fentry.pos.Equals(position) && fentry.size <= int64(maxsize) {
			hitmissCounter.WithLabelValues("envelope", "default", "read", "hit").Inc()
			return *fentry.fragment, fentry.nextPos, nil
		}
	}
	hitmissCounter.WithLabelValues("envelope", "default", "read", "miss").Inc()
	return s.backend.ReadEnvelope(id, position, maxsize)
}

// needCleanup removes the oldest entries until size is 2/3 of the maxsize
// it is a sync function but is should be a background task
func (s *EnvelopeStorage) needCleanup() {
	s.entriesM.Lock()
	defer s.entriesM.Unlock()

	if len(s.entries) <= 1 {
		return
	}

	var (
		targetSize = s.maxsize * 2 / 3
	)

	if atomic.LoadInt64(&s.size) < targetSize {
		return
	}
	sort.Slice(s.entries, func(i, j int) bool {
		return s.entries[i].aTime.Before(s.entries[j].aTime)
	})

	for i, entry := range s.entries {
		newsize := atomic.AddInt64(&s.size, -entry.size)
		s.entriesMap[entry.validator.EnvelopeID] = nil
		delete(s.entriesMap, entry.validator.EnvelopeID)
		s.entries[i] = nil
		sizeInBytesGauge.WithLabelValues("envelope", "default").Set(float64(newsize))
		if newsize <= targetSize || len(s.entries) == 1 {
			s.entries = s.entries[i+1:]
			break
		}
	}
	sizeGauge.WithLabelValues("envelope", "default").Set(float64(len(s.entries)))
}
