package cache

import (
	"testing"

	"github.com/stretchr/testify/assert"
	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/storage"
	"xbus.io/xbus/storage/sql"
	"xbus.io/xbus/testutils"
	"xbus.io/xbus/testutils/mocks"
	"xbus.io/xbus/testutils/sqldb"
	storagetest "xbus.io/xbus/testutils/storage"
)

func envelopeStorageNGFactoryMaker(tb testing.TB) (func(tb testing.TB) (storage.EnvelopeStorage, func()), func()) {
	tb.Helper()

	testDB := sqldb.GetTestDB(tb)
	if testDB == nil {
		tb.Fatal("Failed to init db")
	}
	m, db := testDB.GetModelDB(1)

	return func(tb testing.TB) (storage.EnvelopeStorage, func()) {
			tb.Helper()

			if _, err := db.Engine.DB().Exec("DELETE FROM envelope_storage"); err != nil {
				tb.Fatal(err)
			}
			if _, err := db.Engine.DB().Exec("DELETE FROM envelope_storage_data"); err != nil {
				tb.Fatal(err)
			}
			return sql.NewEnvelopeStorageNG(m, db, testutils.GetTestLogger(tb)), func() {
			}
		}, func() {
			db.Close()
			testDB.Close()
		}
}

func envelopeStorageFactoryMaker(tb testing.TB) (func(tb testing.TB) (storage.EnvelopeStorage, func()), func()) {
	tb.Helper()

	testDB := sqldb.GetTestDB(tb)
	if testDB == nil {
		return nil, nil
	}
	m, db := testDB.GetModelDB(1)

	return func(tb testing.TB) (storage.EnvelopeStorage, func()) {
			tb.Helper()

			if _, err := db.Engine.DB().Exec("DELETE FROM envelope"); err != nil {
				tb.Fatal(err)
			}
			if _, err := db.Engine.DB().Exec("DELETE FROM event"); err != nil {
				tb.Fatal(err)
			}
			if _, err := db.Engine.DB().Exec("DELETE FROM item"); err != nil {
				tb.Fatal(err)
			}

			return sql.NewEnvelopeStorage(m, db, *testutils.GetTestLogger(tb).Zero()), func() {
			}
		}, func() {
			db.Close()
			testDB.Close()
		}
}

func envelopeStorageNGCachedFactoryMaker(tb testing.TB) (func(tb testing.TB) (storage.EnvelopeStorage, func()), func()) {
	tb.Helper()

	factory, clean := envelopeStorageNGFactoryMaker(tb)

	return func(tb testing.TB) (storage.EnvelopeStorage, func()) {
		tb.Helper()

		storage, clean := factory(tb)
		return NewEnvelopeStorage(storage, 0, *testutils.GetTestLogger(tb).Zero()), clean
	}, clean
}

func envelopeStorageCachedFactoryMaker(tb testing.TB) (func(tb testing.TB) (storage.EnvelopeStorage, func()), func()) {
	tb.Helper()

	factory, clean := envelopeStorageFactoryMaker(tb)

	return func(tb testing.TB) (storage.EnvelopeStorage, func()) {
		tb.Helper()

		storage, clean := factory(tb)
		return NewEnvelopeStorage(storage, 0, *testutils.GetTestLogger(tb).Zero()), clean
	}, clean
}

func mockFactory(tb testing.TB) (storage.EnvelopeStorage, func()) {
	tb.Helper()

	return NewEnvelopeStorage(
		mocks.NewEnvelopeStorage(testutils.GetTestLogger(tb)),
		0, *testutils.GetTestLogger(tb).Zero()), func() {}
}

func TestEnvelopeStorage(t *testing.T) {
	t.Run("NGCached", func(t *testing.T) {
		factory, clean := envelopeStorageNGCachedFactoryMaker(t)
		defer clean()

		storagetest.EnvelopeStorageTests(t, factory, storagetest.EnvelopeStorageTestsOptions{
			SupportUnordered:     true,
			SupportParallelStore: true,
		})
	})
	t.Run("LegacyCached", func(t *testing.T) {
		factory, clean := envelopeStorageCachedFactoryMaker(t)
		defer clean()

		storagetest.EnvelopeStorageTests(t, factory, storagetest.EnvelopeStorageTestsOptions{
			SupportUnordered:     false,
			SupportParallelStore: false,
		})
	})
	t.Run("Mock", func(t *testing.T) {
		storagetest.EnvelopeStorageTests(t, mockFactory, storagetest.EnvelopeStorageTestsOptions{
			SupportUnordered:     true,
			SupportParallelStore: true,
		})
	})

	t.Run("cache-size", func(t *testing.T) {
		log := testutils.GetTestLogger(t)
		es := NewEnvelopeStorage(
			mocks.NewEnvelopeStorage(log),
			2000, *log.Zero())

		data := make([]byte, 400)
		for i := 0; i < len(data); i++ {
			data[i] = byte(i)
		}

		envID := api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df")
		envID2 := api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287de")
		envID3 := api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287dd")
		envID4 := api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287dc")
		eventID := api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882")

		env := api.Envelope{
			ID:       envID,
			EventIDs: []api.UUID{eventID},
			Events: []api.Event{
				{ID: eventID, Index: 1, Type: "xbus.test", Items: []api.Item{data}},
			},
		}

		storeEnv := func() {
			state, err := es.StoreEnvelope(env)
			assert.NoError(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, state.Status)

			env.Events[0].Index++

			state, err = es.StoreEnvelope(env)
			assert.NoError(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, state.Status)
		}

		// store a first envelope, only the first fragment will lend in cache
		storeEnv()

		assert.Equal(t, 1, len(es.entriesMap[envID].fragmentList))

		// store a 2 more envelopes, only their first fragment will lend in cache,
		// and the first envelope remains in cache
		env.ID = envID2
		env.Events[0].Index = 1
		storeEnv()

		env.ID = envID3
		env.Events[0].Index = 1
		storeEnv()

		assert.Equal(t, 1, len(es.entriesMap[envID].fragmentList))
		assert.Equal(t, 1, len(es.entriesMap[envID2].fragmentList))
		assert.Equal(t, 1, len(es.entriesMap[envID3].fragmentList))

		// store a last envelope, only the first fragment will lend in cache,
		// and the 2 first envelopes should be removed from the cache so the
		// size goes under 2/3 of the maxsize
		env.ID = envID4
		env.Events[0].Index = 1

		storeEnv()

		assert.NotContains(t, es.entriesMap, envID)
		assert.NotContains(t, es.entriesMap, envID2)
		if assert.Contains(t, es.entriesMap, envID3) {
			assert.Equal(t, 1, len(es.entriesMap[envID3].fragmentList))
		}
		if assert.Contains(t, es.entriesMap, envID4) {
			assert.Equal(t, 1, len(es.entriesMap[envID4].fragmentList))
		}
	})
}

func BenchmarkEnvelopeStorage(b *testing.B) {
	b.Run("NGCached", func(b *testing.B) {
		factory, clean := envelopeStorageNGCachedFactoryMaker(b)
		defer clean()

		storagetest.EnvelopeStorageBench(b, factory)
	})
	b.Run("LegacyCached", func(b *testing.B) {
		factory, clean := envelopeStorageCachedFactoryMaker(b)
		defer clean()

		storagetest.EnvelopeStorageBench(b, factory)
	})
	b.Run("Mock", func(b *testing.B) {
		storagetest.EnvelopeStorageBench(b, mockFactory)
	})
}
