// Copyright 2016 Orus.io. All rights reserved.

package sql

import (
	"context"
	"database/sql"
	"encoding/json"

	"github.com/orus-io/yago"
	"github.com/rs/zerolog"
	"github.com/slicebit/qb"
)

// RawSQLBase provides a base type for implementing pure-sql storages
type RawSQLBase struct {
	db     *yago.DB
	logger zerolog.Logger
}

// MustBegin begins a transaction or panics
func (sb *RawSQLBase) MustBegin() *RawSQLTxHelper {
	return RawSQLMustBegin(sb.db, sb.logger)
}

// RawSQLTxHelper wraps a SQL tx and make it easier to enforce commit or rollback
type RawSQLTxHelper struct {
	*sql.Tx
	complete bool
	logflags qb.LogFlags
	logger   zerolog.Logger
}

// RawSQLMustBegin starts a sql transaction and wrap it
func RawSQLMustBegin(db *yago.DB, logger zerolog.Logger) *RawSQLTxHelper {
	tx, err := db.Engine.DB().Begin()
	if err != nil {
		logger.Err(err).Msg("Cannot start a transaction")
		panic(err)
	}
	return &RawSQLTxHelper{tx, false, db.Engine.Logger().LogFlags(), logger}
}

// Commit commits the transaction
func (tx *RawSQLTxHelper) Commit() error {
	err := tx.Tx.Commit()
	tx.complete = true
	if err != nil {
		tx.logger.Err(err).Msg("Error on commit")
	}
	return err
}

// MustCommit commits of rollback
func (tx *RawSQLTxHelper) MustCommit() {
	if err := tx.Commit(); err != nil {
		panic(err)
	}
}

// Rollback rollbacks and log error if any
func (tx *RawSQLTxHelper) Rollback() error {
	err := tx.Tx.Rollback()
	tx.complete = true
	if err != nil {
		tx.logger.Err(err).Msg("Error on rollback")
	}
	return err
}

// BlindRollback rollbacks the transaction, log errors if any, and returns nothing
func (tx *RawSQLTxHelper) BlindRollback() {
	_ = tx.Rollback()
}

// RollbackIfOpened rollback the transaction is still open
func (tx *RawSQLTxHelper) RollbackIfOpened() {
	if !tx.complete {
		_ = tx.Rollback()
	}
}

func (tx *RawSQLTxHelper) Exec(query string, args ...interface{}) (sql.Result, error) {
	TraceQuery(tx.logflags, &tx.logger, query, args)
	return tx.Tx.Exec(query, args...)
}

func (tx *RawSQLTxHelper) ExecContext(ctx context.Context, query string, args ...interface{}) (sql.Result, error) {
	TraceQuery(tx.logflags, &tx.logger, query, args)
	return tx.Tx.ExecContext(ctx, query, args...)
}

func (tx *RawSQLTxHelper) Query(query string, args ...interface{}) (*sql.Rows, error) {
	TraceQuery(tx.logflags, &tx.logger, query, args)
	return tx.Tx.Query(query, args...)
}

func (tx *RawSQLTxHelper) QueryContext(ctx context.Context, query string, args ...interface{}) (*sql.Rows, error) {
	TraceQuery(tx.logflags, &tx.logger, query, args)
	return tx.Tx.QueryContext(ctx, query, args...)
}

func (tx *RawSQLTxHelper) QueryRow(query string, args ...interface{}) *sql.Row {
	TraceQuery(tx.logflags, &tx.logger, query, args)
	return tx.Tx.QueryRow(query, args...)
}

func (tx *RawSQLTxHelper) QueryRowContext(ctx context.Context, query string, args ...interface{}) *sql.Row {
	TraceQuery(tx.logflags, &tx.logger, query, args)
	return tx.Tx.QueryRowContext(ctx, query, args...)
}

// TraceQuery logs a sql query and args as TRACE level
func TraceQuery(flags qb.LogFlags, log *zerolog.Logger, sql string, args []interface{}) {
	if log != nil && flags != qb.LDefault && log.GetLevel() <= zerolog.DebugLevel {
		e := log.Debug()
		if flags&qb.LQuery != 0 {
			e = e.Str("query", sql)
		}
		if flags&qb.LBindings != 0 {
			arr := zerolog.Arr()
			for _, arg := range args {
				b, err := json.Marshal(arg)
				if err != nil {
					arr.Interface(arg)
				} else {
					if len(b) > 512 {
						b, _ = json.Marshal(map[string]interface{}{
							"arg first 512 bytes": b[:512],
							"len":                 len(b),
						})
					}
					arr.RawJSON(b)
				}
			}
			e = e.Array("args", arr)
		}
		e.Msg("SQL Query")
	}
}
